# -*- coding: utf-8 -*-

"""
PokeAPI.co API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the PokeAPI.co API.
:copyright: (c) 2019 Lorenzo
:license: MIT, see LICENSE for more details.
"""

__title__ = 'async_pokepy'
__author__ = 'Lorenzo'
__license__ = 'MIT'
__copyright__ = 'Copyright 2019 Lorenzo'
__version__ = '0.0.9a'

from .client import Client
from .exceptions import *
from .types import *
